/*
 * This file is part of the OWL API.
 *
 * The contents of this file are subject to the LGPL License, Version 3.0.
 *
 * Copyright (C) 2011, The University of Manchester
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 *
 * Alternatively, the contents of this file may be used under the terms of the Apache License, Version 2.0
 * in which case, the provisions of the Apache License Version 2.0 are applicable instead of those above.
 *
 * Copyright 2011, University of Manchester
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package uk.ac.manchester.cs.owlapi.modularity;

/** Enumeration for module types.
 * 
 * @author Thomas Schneider
 * @author School of Computer Science
 * @author University of Manchester */
public enum ModuleType {
    /** Module type representing top modules (aka lower modules). */
    TOP("top"),
    /** Module type representing bottom modules (aka upper modules). */
    BOT("bottom"),
    /** @deprecated Module type representing bottom-of-top modules (aka
     *             upper-of-lower modules). */
    @Deprecated
    BOT_OF_TOP("bottom-of-top"),
    /** @deprecated Module type representing top-of-bottom modules (aka
     *             lower-of-upper modules). */
    @Deprecated
    TOP_OF_BOT("top-of-bottom"),
    /** Module type representing fixpoint-nested modules
     * (top-of-bottom-of-top-of-...). */
    STAR("nested");
    /** The name of the module type in bottom/top notation. */
    private final String name;

    /** Constructs a module type with the given name and alternative name.
     * 
     * @param name
     *            the name in bottom/top notation */
    ModuleType(String name) {
        this.name = name;
    }

    @Override
    public String toString() {
        return name;
    }
}
